#
#  Copyright (C) 2012, 2013, 2015
#            Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Find ancillary files from a file header (for Chandra data).

The current list of header keyword and file types are:

  Keyword     Filetype

  ASOLFILE    asol
  BPIXFILE    bpix
  DTFFILE     dtf
  FLTFILE     flt
  MASKFILE    mask
  MTLFILE     mtl
  PBKFILE     pbk

The following keywords may be supported at a later date.

  THRFILE     thresh
  GAINFILE    gain
  CTIFILE     cti
  GRD_FILE    grade
  AIMPFILE    aimpoint
  GEOMFILE    geom
  SKYFILE     sky
  TDETFILE    tdet

Examples:

>>> find_ancillary_files('50/repro/acisf00050_repro_evt2.fits', ['asol', 'mtl', 'pbk', 'mask'], absolute=False)
[['50/repro/pcadf053035261N003_asol1.fits'],
 ['NONE'],
 ['50/repro/acisf053036614N003_pbk0.fits'],
 ['50/repro/acisf00050_000N003_msk1.fits']]
>>> find_ancillary_files('50/repro/acisf00050_repro_evt2.fits', ['asol', 'mtl', 'pbk', 'mask'], absolute=True)
[['/example/path/50/repro/pcadf053035261N003_asol1.fits'],
 ['NONE'],
 ['/example/path/50/repro/acisf053036614N003_pbk0.fits'],
 ['/example/path/50/repro/acisf00050_000N003_msk1.fits']]

Notes:

  - as of CIAO 4.6 there should be no need for the PBK (parameter block) file,
    but support has been left in.

"""

import os

import pycrates
import stk

from ciao_contrib.logger_wrapper import initialize_module_logger

__all__ = (
    'find_ancillary_files',
    'find_ancillary_files_crate',
    'find_ancillary_files_header'
    )

lgr = initialize_module_logger('ancillaryfiles')
v1 = lgr.verbose1
v3 = lgr.verbose3
v4 = lgr.verbose4

__typemap = {
    'asol': 'ASOLFILE',
    'bpix': 'BPIXFILE',
    'dtf': 'DTFFILE',
    'flt': 'FLTFILE',
    'mask': 'MASKFILE',
    'mtl': 'MTLFILE',
    'pbk': 'PBKFILE',
    # 'thresh': 'THRFILE',
    # 'gain': 'GAINFILE',
    # 'cti': 'CTIFILE',
    # 'grade': 'GRD_FILE',
    # 'aimpoint': 'AIMPFILE',
    # 'geom': 'GEOMFILE',
    # 'sky': 'SKYFILE',
    # 'tdet': 'TDETFILE',
    }


def validate_filetype(ftype):
    """Returns the header value for the
    given file type. Raises a ValueError if ftype
    is invalid. ftype is case insensitive.
    """

    try:
        key = __typemap[ftype.lower()]

    except KeyError:
        tkeys = __typemap.keys()
        tkeys.sort()
        raise ValueError("Invalid filetype '{}', ".format(ftype) +
                         "must be one of: {}".format(" ".join(tkeys)))

    return key


def find_file(filename, absolute=True, cwd='.'):
    """Return the path to the given file or None if it can
    not be found. We assume that filename contains no DM filter (or
    if it does then it doesn't contain any / characters within the filter
    that would confuse os.path.abspath).

    See find_ancillary_files_header() for a discussion of the search
    path and description of the absolute and cwd arguments.
    """

    search_path = ['.',
                   'repro', 'primary', 'secondary',
                   '../repro', '../primary', '../secondary',
                   ]

    base = os.path.abspath(cwd)
    v4("Looking for {} with base={}".format(filename, base))
    for p in search_path:
        fullpath = os.path.normpath(os.path.join(base, p, filename))
        v4(" -> " + fullpath)
        if os.path.exists(fullpath) or os.path.exists(fullpath + '.gz'):
            v4("    found")
            if absolute:
                return fullpath
            else:
                return os.path.relpath(fullpath, start='.')

    v4("File not found.")
    return None


def find_files(fnames, absolute=True, cwd='.', keepnone=True):
    """Return an array of file names giving the absolute (absolute=True)
    or relative path to fnames, if they are all found, otherwise None is
    returned.

    fnames is a string and is treated as a stack.

    cwd gives the location of the directory to use as the base when
    calculating relative paths (so it is only used if absolute=False).

    If keepnone is True then a value of NONE is retained, otherwise
    the return value is None rather than an array. This lets you
    identify when a file value was given as NONE versus it not
    being found.
    """

    out = []
    v4("Calling find_files on " + fnames)
    for fname in stk.build(fnames):
        if fname == 'NONE':
            if keepnone:
                out.append(fname)
                continue
            else:
                return None

        path = find_file(fname, absolute=absolute, cwd=cwd)
        if path is None:
            return None

        out.append(path)

    return out


def find_ancillary_files_header(header, filetypes, absolute=True,
                                cwd='.', keepnone=True):
    """Given a file header (dictionary of the header keywords)
    and a list of file types, return the paths to these
    files.

    The return is an array, in the same order as the filetypes
    array. Each element is either an array of file names, or None.  In
    most cases the array will be a single element (aspect solutions
    can have multiple files). For the case where there are multiple
    files, all must be found, otherwise None is returned.

    The header keywords can be set to 'NONE': the return value for
    this case depends on the keepnone keyword. A value of False means
    that None is returned, otherwise the string 'NONE' is returned.
    This is intended to allow users to distinguish from 'there is no
    file set' vs 'the file is missing or can not be found'.

    Valid file types are (case insensitive):

      asol bpix dtf flt mask mtl pbk

    A ValueError is raised if an unknown file type is used.

    If absolute is True then the absolute path is given, otherwise a
    relative path is returned. When calculating the relative version,
    the current working directory is used as the bath (i.e.  not the
    value of the cwd argument given to this routine).

    The search for each file is made (relative to the cwd
    directory):

        .
        repro/
        primary/
        secondary/
        ../repro/
        ../primary/
        ../secondary/

    Behavior with gzipped files:

    The file name is returned even if it doesn't exist but there
    is a *.gz version present. This is supported by most CIAO
    tools transparently, but there are places where it may not
    work (in particular the ardlib settings should use the
    actual file name).

    """

    out = []
    for ftype in filetypes:
        key = validate_filetype(ftype)

        try:
            fnames = header[key]

        except KeyError:
            out.append(None)
            continue

        except TypeError:
            raise ValueError("The header argument does not appear to be a dictionary.")

        v3("Header: {} = {}".format(ftype, fnames))

        tout = find_files(fnames, absolute=absolute, cwd=cwd,
                          keepnone=keepnone)
        out.append(tout)

    return out


def find_ancillary_files_crate(cr, filetypes, absolute=True,
                               cwd='.', keepnone=True):
    """Find the location of the given ancillary files from the header in
    the crate cr.

    This is a wrapper around find_ancillary_files_header(); please see the
    help for that routine for a description of all the arguments
    except cr.

    """

    header = {}

    def set_header(k):
        v = cr.get_key_value(k)
        if v is not None:
            header[k] = v

    for ftype in filetypes:
        key = validate_filetype(ftype)
        set_header(key)

    v4("find_ancillary_files_crate: header={}".format(header))
    return find_ancillary_files_header(header,
                                       filetypes,
                                       absolute=absolute,
                                       cwd=cwd,
                                       keepnone=keepnone)


def find_ancillary_files(filename, filetypes, absolute=True, keepnone=True):
    """Find the location of the given ancillary files (defined
    by the filetypes array) from the header in
    the file called filename.

    The return is an array, in the same order as the filetypes
    array. Each element is either an array of file names, or None.  In
    most cases the array will be a single element (aspect solutions
    can have multiple files). For the case where there are multiple
    files, all must be found, otherwise None is returned.

    The header keywords can be set to 'NONE': the return value for
    this case depends on the keepnone keyword. A value of False means
    that None is returned, otherwise the string 'NONE' is returned.
    This is intended to allow users to distinguish from 'there is no
    file set' vs 'the file is missing or can not be found'.

    Valid file types are (case insensitive):

      asol bpix dtf flt mask mtl pbk

    A ValueError is raised if an unknown file type is used.

    If absolute is True then the absolute path is given, otherwise a
    relative path is returned. When calculating the relative version,
    the current working directory is used as the bath.

    The search for each file is made in the following locations,
    relative to the location of filename:

        .
        repro/
        primary/
        secondary/
        ../repro/
        ../primary/
        ../secondary/

    Behavior with gzipped files:

    The file name is returned even if it doesn't exist but there
    is a *.gz version present. This is supported by most CIAO
    tools transparently, but there are places where it may not
    work (in particular the ardlib settings should use the
    actual file name).

    Examples:

    >>> find_ancillary_files('50/repro/acisf00050_repro_evt2.fits', ['asol', 'mtl', 'pbk', 'mask'], absolute=False)
    [['50/repro/pcadf053035261N003_asol1.fits'],
     ['NONE'],
     ['50/repro/acisf053036614N003_pbk0.fits'],
     ['50/repro/acisf00050_000N003_msk1.fits']]
    >>> find_ancillary_files('50/repro/acisf00050_repro_evt2.fits', ['asol', 'mtl', 'pbk', 'mask'], absolute=True)
    [['/example/path/50/repro/pcadf053035261N003_asol1.fits'],
     ['NONE'],
     ['/example/path/50/repro/acisf053036614N003_pbk0.fits'],
     ['/example/path/50/repro/acisf00050_000N003_msk1.fits']]

    """

    # Remove DM filter since it can mess up the os.path routines.
    # This is not robust but I am willing to live with the
    # assumption that users do not have many files with [ as a
    # valid character.
    #
    lpos = filename.find('[')
    if lpos == -1:
        fullname = os.path.normpath(os.path.abspath(filename))
    else:
        fullname = os.path.normpath(os.path.abspath(filename[:lpos]))

    basedir = os.path.dirname(fullname)
    cr = pycrates.read_file(filename)
    out = find_ancillary_files_crate(cr, filetypes,
                                     absolute=absolute,
                                     cwd=basedir,
                                     keepnone=keepnone)
    return out
