#
#  Copyright (C) 2009, 2010, 2011, 2012, 2014, 2015
#            Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Utility routines for Sherpa.

  get_instmap_weights()
  save_instmap_weights()
  plot_instmap_weights()
  estimate_weighted_expmap()

"""

import os

import logging
import time

import numpy as np

from itertools import izip

# from sherpa.astro.ui import *
from sherpa.astro import ui
from sherpa.astro.utils import _charge_e
import sherpa.utils as su

from sherpa.utils import print_fields
from sherpa.data import Data1DInt
from sherpa.astro.data import DataPHA
from sherpa.models.parameter import Parameter

import pychips.all as pyc
import pycrates

# from crates_contrib.utils import make_table_crate

__all__ = ["get_instmap_weights", "save_instmap_weights",
           "plot_instmap_weights", "estimate_weighted_expmap",
           "InstMapWeights", "InstMapWeights1DInt", "InstMapWeightsPHA"]

logger = logging.getLogger("sherpa")


def _is_boolean(val):
    "Returns True if val is a boolean."
    return isinstance(val, (bool, np.bool_))


class InstMapWeights:
    """Store the weights information needed by mkinstmap.

    Attributes are:

      id   - dataset id
      modelexpr - string representation of the source model
      xmid - mid-point of each bin
      xlo  - lower edge of ech bin
      xhi  - upper edge of each bin
      weight - weight value for each bin
      fluxtype - weights are to create an instrument map with
                 units of  cm^2 count / <fluxtype>
                 Can be 'photon' or 'erg'.

    It is up to the user to ensure that the model evaluates to
    photon/cm^2/s when integrated over a bin and that the bins are in
    keV.
    """

    _valid_fluxtypes = ["erg", "photon"]

    def __str__(self):
        def fmt(key, val):
            return "{0:10} = {1}".format(key, str(val))

        names = ["id", "modelexpr", "xlo", "xhi", "xmid", "weight", "fluxtype"]
        vals  = [getattr(self, n) for n in names]
        return print_fields(names, dict(izip(names, vals)))

    def __init__(self, id=None, fluxtype="photon"):
        "If id is None the default id will be used."

        if id is None:
            self.id = ui.get_default_id()
        else:
            self.id = id

        if fluxtype in self._valid_fluxtypes:
            self.fluxtype = fluxtype
        else:
            raise ValueError("fluxtype set to {0} but must be one of: {1}".format(
                fluxtype, " ".join(self._valid_fluxtypes)))

        # Set up the xlo/xhi/xmid arrays
        d = ui.get_data(self.id)
        self._calc_bins(d)
        self._apply_mask(d)

        # Important to use get_source and not get_model as we do not
        # want to apply any instrument model to the evaluation.
        #
        # Note that we do not hold onto the model expression object,
        # which is probably not an issue here.
        #
        mdl = ui.get_source(id)
        self.modelexpr = mdl.name

        # We do not use xlo/xhi but the _xlo/_xhi attributes which
        # contain an extra bin, in case of X-Spec models
        #
        src = mdl(self._xlo, self._xhi)[:-1]
        if np.any(src < 0.0):
            raise RuntimeError("There are negative values in your source model (id={0})!".format(self.id))
        if np.all(src <= 0.0):
            raise RuntimeError("The source model for id={0} evaluates to 0!".format(self.id))

        # Conversion to a single datatype is a bit excessive here.
        #
        dtype = src.dtype

        if self.fluxtype == "erg":
            norm = _charge_e * np.sum(src * self.xmid)
        else:
            norm = np.sum(src)

        self.weight = src / norm

        self.weight = self.weight.astype(dtype)
        self.xlo    = self.xlo.astype(dtype)
        self.xhi    = self.xhi.astype(dtype)
        self.xmid   = self.xmid.astype(dtype)

    def _calc_bins(self, data):
        "Calculate the bin edges"

        raise NotImplementedError("_calc_bins")

    def _apply_mask(self, data):
        "Apply any mask/filter"

        # Apply the filter if necessary.
        #
        m = data._get_mask()
        if _is_boolean(m):
            if not m:
                raise RuntimeError("filter excludes all data")
        else:
            self.xlo = self.xlo[m]
            self.xhi = self.xhi[m]

        if len(self.xlo) == 0:
            raise RuntimeError("filter excludes all data")
        if len(self.xlo) == 1:
            raise RuntimeError("Only 1 data point in grid!")

        self.xmid = 0.5 * (self.xlo + self.xhi)

        # Add in an extra point in case dealing with X-Spec models
        # (where the last bin is used to define the bin edges, so
        # the model evaluates to 0 there).
        #
        # Errr, given that we explicitly give the low and high bin
        # edges to the models, I think this is actually unnecessary.
        #
        self._xlo = np.append(self.xlo,
                              [self.xhi[-1]])
        self._xhi = np.append(self.xhi,
                              [self.xhi[-1] + self.xlo[1] + self.xlo[0]])

    def save(self, filename, clobber=True):
        """Save the weights to filename as an ASCII file, that
        can be used as the spectrumfile input for mkinstmap.

        If filename already exists it will be overwritten unless
        clobber is False, in which case a warning will be thrown.
        """

        # Used to use crates, but it does not write out keywords for
        # the TEXT/SIMPLE format, which we now want. So we now manually
        # create the output file.
        #
        # cr = make_table_crate(self.xmid, self.weight,
        #      colnames=["X", "WEIGHT"])
        # key = cr.get_key("CREATOR")
        # key.value = self.__class__.__name__
        # cr.write("{0}[opt kernel=text/simple]".format(filename),
        #          clobber=clobber)

        if not clobber and os.path.exists(filename):
            raise IOError("{} exists and clobber=False".format(filename))

        elo = self.xlo[0]
        ehi = self.xhi[-1]

        stime = time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime())

        # The column names are written first so that the file can easily be
        # used with CIAO tools (as they assume 'opt colname=first' by
        # default).
        #
        with open(filename, 'w') as fh:
            fh.write('#TEXT/SIMPLE\n')
            fh.write('# X WEIGHT\n')
            fh.write('#\n')
            fh.write("# DATE = {} / Date and time of file creation\n".format(stime))
            fh.write("# MODELEXPR = {}\n".format(self.modelexpr))
            fh.write("# ENERG_LO = {} / [keV] Minimum energy\n".format(elo))
            fh.write("# ENERG_HI = {} / [keV] Maximum energy\n".format(ehi))
            fh.write('#\n')
            for (x, weight) in zip(self.xmid, self.weight):
                # As we expect weight values to be ~ 0.01 to 1
                # and the energy values to also be "nice" then
                # rely on the default formatting rules used by
                # Python.
                fh.write("{} {}\n".format(x, weight))

        logger.info("Created: {}".format(filename))

    def plot(self, overplot=False):
        """Plot the weights values.

        The data plot preferences are used to control the
        appearance of the plot: at present the following fields
        are used:
          xlog and ylog
          linethickness, linecolor

          (linestyle is always set to chips_solid)

        If replot is True then the data is added to the current
        plot.
        """

        prefs = ui.get_data_plot_prefs()

        # There is no validation of the preference values
        #
        hprefs = pyc.ChipsHistogram()
        hprefs.symbol.style = pyc.chips_none
        if prefs.get("linethickness", None) is not None:
            hprefs.line.thickness = prefs["linethickness"]
        if prefs.get("linecolor", None) is not None:
            hprefs.line.color = prefs["linecolor"]
        # if prefs.get("linestyle", None) is not None:
        #     hprefs.line.style = prefs["linestyle"]
        hprefs.line.style = pyc.chips_solid

        # We rely on XY_AXIS == X_AXIS | Y_AXIS
        #
        xlog = prefs.get("xlog", False)
        ylog = prefs.get("ylog", False)
        axis = 0
        if xlog: axis |= pyc.X_AXIS
        if ylog: axis |= pyc.Y_AXIS

        pyc.open_undo_buffer()
        try:

            if not overplot:
                pyc.erase()

            # I would like to use self.xlo/self.xhi here but this can
            # lead to ugly plots for certain datasets created by dataspace1d.
            # So we use xmid for now. This may no-longer be an issue
            # (CIAO 4.2), but left as-is for now.
            #
            pyc.add_histogram(self.xmid, self.weight, hprefs)

            if not overplot:
                pyc.set_plot_xlabel("Energy (keV)")
                pyc.set_plot_ylabel("Weights")
                pyc.set_plot_title("Weights for {0}: {1}".format(self.id, self.modelexpr))

                if axis > 0:
                    pyc.log_scale(axis)

        except:
            pyc.discard_undo_buffer()
            raise

        pyc.close_undo_buffer()

    def _estimate_expmap(self, *args):
        """Estimate the exposure map given an ARF.

        The input arguments can be one of

          filename    - name of an ARF
          crate       - uses the energ_lo, energ_hi, and specresp columns
          xlo, xhi, y - arrays taken to be the energ_lo, energ_hi, and
                        specresp columns

        The output is an estimate of the exposure map at the position
        of the source, and has units of cm^2 count / <fluxtype>.

        The ARF is linearly interpolated onto the energy grid
        of the dataset and the weighted sum calculated. The
        ARF is assumed to have units of cm^2 count / photon, and
        be defined on a grid given in keV.

        The estimate_instmap() method should be called in
        preference to this routine.
        """

        nargs = len(args)
        if nargs == 3:
            elo = args[0]
            ehi = args[1]
            specresp = args[2]

            if len(elo) == 1 or len(ehi) == 1 or len(specresp) == 1:
                raise TypeError("Expected three arrays of the same length, with more than one element in")
            if len(elo) != len(ehi) or len(elo) != len(specresp):
                raise ValueError("Expected three arrays of the same length")

        elif nargs != 1:
            raise TypeError("_estimate_expmap() takes 2 or 4 arguments ({0} given)".format(nargs + 1))

        else:

            if isinstance(args[0], pycrates.TABLECrate):
                cr = args[0]
            else:
                cr = pycrates.TABLECrate(args[0], mode="r")

            cnames = ["ENERG_LO", "ENERG_HI", "SPECRESP"]
            if any([not cr.column_exists(x) for x in cnames]):
                raise ValueError("Crate '{0}' does not contain one of the columns: {1}".format(
                                 cr.get_filename(), " ".join(cnames)))

            elo = cr.get_column("ENERG_LO").values.copy()
            ehi = cr.get_column("ENERG_HI").values.copy()
            specresp = cr.get_column("SPECRESP").values.copy()

        # Interpolate using the mid-point of the ARF
        # onto the mid-point of the grid if necessary.
        #
        emid = 0.5 * (elo + ehi)
        if np.all(emid == self.xmid):
            arf = specresp
        else:
            arf = [su.interpolate(x, emid, specresp) for x in self.xmid]
            arf = np.asarray(arf, dtype=self.weight.dtype)
            arf[arf < 0] = 0.0

        return np.sum(self.weight * arf)

    def estimate_expmap(self, *args):
        """Estimate the exposure map given an ARF.

        The input arguments can be one of

          filename    - name of an ARF
          crate       - uses the energ_lo, energ_hi, and specresp columns
          xlo, xhi, y - arrays taken to be the energ_lo, energ_hi, and
                        specresp columns

        The output is an estimate of the exposure map at the position
        of the source, and has units of cm^2 count / <fluxtype>.
        """

        nargs = len(args)
        if nargs != 1 and nargs != 3:
            raise TypeError("estimate_expmap() takes 2 or 4 arguments ({0} given)".format(nargs + 1))
        return self._estimate_expmap(*args)


class InstMapWeights1DInt(InstMapWeights):
    "Instrument map weights for Data1DInt datasets"

    def _calc_bins(self, data):
        "Get the xlo/xhi bin values"

        self.xlo = data.xlo

        # Handle a case when dataspace1d in CIAO 4.1.2 can produce different
        # length bins. If this has happened then we know that d._get_mask()
        # will be True since ignore/notice will fail to work on such a
        # dataset. We do not take advantage of this knowledge since it doesn't
        # really help.
        #
        nlo = len(data.xlo)
        nhi = len(data.xhi)
        if nlo == nhi:
            self.xhi = data.xhi

        elif nhi == nlo + 1:
            self.xhi = data.xhi[:-1]

        else:
            raise RuntimeError("Internal error: need to handle .xlo/xhi lengths of {0} and {1}!".format(nlo, nhi))


class InstMapWeightsPHA(InstMapWeights):
    "Instrument map weights for DataPHA datasets"

    def _calc_bins(self, data):
        "Get the xlo/xhi bin values"

        (self.xlo, self.xhi) = data._get_ebins()

    def estimate_expmap(self, *args):
        """Estimate the exposure map. The input arguments
        can be one of

          none        - use the ARF of the associated dataset
          filename    - name of an ARF
          crate       - uses the energ_lo, energ_hi, and specresp columns
          xlo, xhi, y - arrays taken to be the energ_lo, energ_hi, and
                        specresp columns

        The output is an estimate of the exposure map at the position
        of the source, and has units of cm^2 count / <fluxtype>.
        """

        nargs = len(args)
        if nargs == 0:
            darf = ui.get_arf(self.id)
            return self._estimate_expmap(darf.energ_lo,
                                         darf.energ_hi,
                                         darf.specresp)
        else:
            return self._estimate_expmap(*args)


def get_instmap_weights(id=None, fluxtype="photon"):
    """Returns an InstMapWeights object with the weights information
    for use by mkinstmap.

    If id is None then the default id is used.

    The fluxtype argument gives the units of the instrument map:
    cm^2 count / <fluxtype>. The valid values are "photon" (the default)
    or "erg". If set to photon then the weights will sum to 1.

    An error will be thrown if the model evaluates to
    a negative value, or there is no flux.

    This is intended for use with a dataset "faked" using

      dataspace1d(elow, ehigh, estep)
      set_source(...)

    although there is an attempt to support DataPHA objects
    (either for spectra that have been loaded in or "faked"
    using dataspace1d, specifying the data type explicitly).
    """

    if id is None:
        id = ui.get_default_id()

    # Since sherpa.astro.data.DataPHA is a subclass of sherpa.data.Data1DInt
    # we need to check for it first.
    #
    d = ui.get_data(id)
    if isinstance(d, DataPHA):
        return InstMapWeightsPHA(id, fluxtype=fluxtype)
    elif isinstance(d, Data1DInt):
        return InstMapWeights1DInt(id, fluxtype=fluxtype)
    else:
        raise RuntimeError("Unable to calculate weights from a dataset of type {0}.{1}".format(
                           d.__class__.__module__, d.__class__.__name__))


def save_instmap_weights(*args, **kwargs):
    """Save a weights file for use by mkinstmap.

    The arguments that can be given without a name are

        filename
        id, filename
        filename, clobber
        id, filename, clobber

    and the only argument that has to be given as a named argument is
    fluxtype (which defaults to "photon"). If the id argument is not
    given then the default id is used, and clobber defaults to True.

    The fluxtype argument gives the units of the instrument map:
    cm^2 count / <fluxtype>. The valid values are "photon" (the default)
    or "erg". If set to photon then the weights will sum to 1.

    The command

      save_instmap_weights("foo.dat", fluxtype="erg", clobber=False)

    saves the weights to the file foo.dat, and will error out if it
    already exists. The weights are written out so as to create an
    instrument map with units of cm^2 count / erg.
    """

    fname = "save_instmap_weights"
    nargs = len(args)
    if nargs == 0:
        raise TypeError("{0}() takes at least 1 argument (0 given)".format(fname))
    if nargs > 3:
        raise TypeError("{0}() takes at most 3 arguments ({1} given)".format(fname, nargs))

    # The default values
    user = {"id": ui.get_default_id(),
            "filename": None,
            "clobber": True,
            "fluxtype": "photon"}
    argnames = user.keys()

    if nargs == 1:
        user["filename"] = args[0]

    elif nargs == 3:
        user["id"]       = args[0]
        user["filename"] = args[1]
        user["clobber"]  = args[2]

    elif _is_boolean(args[1]):
        user["filename"] = args[0]
        user["clobber"]  = args[1]

    elif isinstance(args[1], int):
        # This was needed in CIAO 4.2 and earlier since S-Lang would
        # end up using integers for boolean values. Left in in CIAO 4.3
        #
        user["filename"] = args[0]
        user["clobber"]  = args[1] == 1

    else:
        user["id"]       = args[0]
        user["filename"] = args[1]

    for (n, v) in kwargs.iteritems():
        if n not in argnames:
            raise TypeError("{0}() got an unexpected keyword argument '{1}'".format(fname, n))
        user[n] = v

    wgts = get_instmap_weights(user["id"], fluxtype=user["fluxtype"])
    wgts.save(user["filename"], clobber=user["clobber"])


def plot_instmap_weights(id=None, overplot=False, fluxtype="photon"):
    """Plot the weights values.

    If id is None then the default id is used.

    The fluxtype argument gives the units of the instrument map:
    cm^2 count / <fluxtype>. The valid values are "photon" (the default)
    or "erg". If set to photon then the weights will sum to 1.

    The data plot preferences are used to control the
    appearance of the plot: at present the following fields
    are used:
      xlog and ylog
      linethickness, linecolor

    (linestyle is always set to chips_solid)

    If replot is True then the data is added to the current
    plot.
    """

    if id is None:
        id = ui.get_default_id()

    wgts = get_instmap_weights(id, fluxtype=fluxtype)
    wgts.plot(overplot=overplot)


def estimate_weighted_expmap(id=None, arf=None, elo=None, ehi=None,
                             specresp=None, fluxtype="photon",
                             par=None, pvals=None):
    """Estimate the weighted exposure map value for an ARF,
    given a spectral model and energy range.

    If par and pvals are None then the return value is a scalar value,
    evaluated using the current model parameters for the appropriate source
    model. If they are not None then the return value is an array of values,
    evaluated for each parameter value in parvals for the parameter given by
    par (all the other model paramter values are left at the current
    values). On exit the original parameter value will be restored. So

      estimate_weighted_expmap(arf="arf.fits", par=pl.gamma,
                               pvals=np.arange(0.5,5,0.1))

    will calcualte the value for the pl.gamma values from 0.5 up to 5,
    with a step size of 0.1.

    If id is None then the default id is used.

    The fluxtype argument gives the units of the return value:
    cm^2 count / <fluxtype>. The valid values are "photon" (the default)
    or "erg".

    The ARF to use is defined in one of several ways:

    a) If the dataset is a PHA file with an associated ARF, leave all
       the arf, elo, ehi or specresp arguments as None and this ARF
       will be used

    b) if the arf is set, then it is assumed to be either the name of
       the ARF to use or be a table crate of the ARF

    c) if arf is None and elo,ehi,specresp are set, then they are taken to
       be the arrays defining the ARF: elo and ehi give the bin edges in
       keV and specresp the value of the ARF in cm^2.

    The ARF is interpolated onto the energy grid of the dataspace.

    """

    # Usage errors. We can not catch them all before doing actual work.
    #
    if arf is None and \
       (elo is not None or ehi is not None or specresp is not None):
        # we only worry about elo/ehi/specresp if arf is NOT given
        if elo is None or ehi is None or specresp is None:
            raise TypeError("Either elo,ehi,specresp must all be set or all be None.")

    if id is None:
        id = ui.get_default_id()

    if par is not None or pvals is not None:

        if par is None or pvals is None:
            raise TypeError("Either both par and pvals are set or they are both None.")

        if not isinstance(par, Parameter):
            raise TypeError("par argument must be a Sherpa model parameter.")

        if not hasattr(pvals, "__iter__"):
            raise TypeError("pvals argument must be an array/list.")

        smdl = ui.get_source(id)
        if par not in smdl.pars:
            raise TypeError("par argument is not a parameter of the source model")

    wgts = get_instmap_weights(id, fluxtype=fluxtype)
    if isinstance(wgts, InstMapWeights1DInt) and arf is None and elo is None:
        raise TypeError("The arf parameter or the elo,ehi,specresp parameters must be given.")

    if arf is None:
        if elo is None:
            args = []
        else:
            args = [elo, ehi, specresp]
    else:
        args = [arf]

    if par is None:
        return wgts.estimate_expmap(*args)

    else:
        # Ugh: we have to create an object for each evaluation, which
        # is rather wasteful.
        #
        orig = par.val
        out = []
        try:
            for pval in pvals:
                par.val = pval
                wgts = get_instmap_weights(id, fluxtype=fluxtype)
                out.append(wgts.estimate_expmap(*args))
        except:
            par.val = orig
            raise

        par.val = orig
        return np.asarray(out)

# End
